#include "lab3d.h"

/* Size of sounds.kzp should be this. Otherwise, panic. */
#define SND_FILE_LENGTH 294352

/* Initialise the game: set up most I/O and data... */

void initialize()
{
    K_INT16 i, j, k, walcounter, oclockspeed;
    K_UINT16 l;
    char *v;
    time_t tnow;
    SDL_Surface *screen;

    SDL_AudioSpec want;
    FILE *file;

    int realr,realg,realb,realz,reald;

    spriteyoffset=0;
    ingame=0;
    mixing=0;
    menuing=0;

    visiblescreenyoffset=0;

    soundmutex=SDL_CreateMutex();
    timermutex=SDL_CreateMutex();

    /* Initialise SDL; uncomment the NOPARACHUTE bit if the parachute
       routine (which catches stuff like segfaults) gets in the way of your
       debugging. */

    SDL_Init(SDL_INIT_TIMER|SDL_INIT_VIDEO/*|SDL_INIT_NOPARACHUTE*/|
	     SDL_INIT_JOYSTICK);
    SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,8);
    SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,8);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,24);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);
    SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE,0);
    SDL_GL_SetAttribute(SDL_GL_ACCUM_RED_SIZE,0);
    SDL_GL_SetAttribute(SDL_GL_ACCUM_GREEN_SIZE,0);
    SDL_GL_SetAttribute(SDL_GL_ACCUM_BLUE_SIZE,0);
    SDL_GL_SetAttribute(SDL_GL_ACCUM_ALPHA_SIZE,0);
    SDL_ShowCursor(0);

    fprintf(stderr,"Activating video...\n");

    if ((screen=SDL_SetVideoMode(screenwidth, screenheight, 32, 
				 fullscreen?
				 (SDL_OPENGL|SDL_FULLSCREEN):SDL_OPENGL))==
	NULL) {
	fprintf(stderr,"True colour failed; taking whatever is available.\n");
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,5);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,5);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,16);
	if ((screen=SDL_SetVideoMode(screenwidth, screenheight, 0, 
				     fullscreen?
				     (SDL_OPENGL|SDL_FULLSCREEN):SDL_OPENGL))==
	    NULL) {
	    fprintf(stderr,"Video mode set failed.\n");
	    SDL_Quit();
	    exit(-1);
	}
    }	

    SDL_GL_GetAttribute(SDL_GL_RED_SIZE,&realr);
    SDL_GL_GetAttribute(SDL_GL_GREEN_SIZE,&realg);
    SDL_GL_GetAttribute(SDL_GL_BLUE_SIZE,&realb);
    SDL_GL_GetAttribute(SDL_GL_DEPTH_SIZE,&realz);
    SDL_GL_GetAttribute(SDL_GL_DOUBLEBUFFER,&reald);

    if (reald==0) {
	fprintf(stderr,"Double buffer not available.\n");
	SDL_Quit();
	exit(-1);
    }

    if (realz<16) {
	fprintf(stderr,"Warning: Depth buffer resolution too low; expect\n");
	fprintf(stderr,"graphical glitches.\n");
	SDL_Quit();
	exit(-1);
    }

    if (realz<24) {
	walltol=256; neardist=128;
    } else {
	walltol=32; neardist=16;
    }

    fprintf(stderr,"Opened GL at %d/%d/%d bits, %d bit depth buffer.\n",
	    realr,realg,realb,realz);

    /* SDL on Windows does funny things if the specified window size or
       fullscreen resolution is invalid. SDL on X fakes it seamlessly. The
       documentation doesn't mention anything about this, so I assume that this
       aspect of SDL is undefined. */

    if ((screenwidth!=screen->w)||(screenheight!=screen->h))
	fprintf(stderr,"Warning: resolution is actually %dx%d.\n",screen->w,
		screen->h);

    screenwidth=screen->w;
    screenheight=screen->h;

    /* Actually, this mode seems to be both more compatible and faster, so
       I think I'll leave it like this. Change this to 1 at your own risk. */

    largescreentexture=0;
							      
    if (largescreentexture) {
	/* One large 512x512 texture. */

	screenbufferwidth=screenbufferheight=512;
    } else {
	/* 6*8 matrix of 64x64 tiles with 1 pixel wide borders on shared
	   edges. */

	screenbufferwidth=374;
	screenbufferheight=498;
    }

    screenbuffer=malloc(screenbufferwidth*screenbufferheight);    
    SDL_WM_SetCaption("Ken's Labyrinth", "Ken's Labyrinth");

    linecompare(479);

    if (screenbuffer==NULL) {
	fprintf(stderr,"Insufficient memory.\n");
	SDL_Quit();
	exit(-1);
    }
    
    fprintf(stderr,"Loading tables/settings...\n");

    loadtables();
    vidmode = 1; /* Force fake 360x240 mode. */
    mute = 0;
    if ((joyx1 == joyx2) || (joyx2 == joyx3) || (joyy1 == joyy2) ||
	(joyy2 == joyy3))
	joystat = 1;

    if (joystat==0) {
        fprintf(stderr,"Opening joystick...\n");
	joystick=SDL_JoystickOpen(0);
    }

    if (joystick==NULL) joystat=1;

    time(&tnow);
    srand((unsigned int)tnow);
    if ((note = convallocate(16384>>4)) == NULL)
    {
	fprintf(stderr,"Error #1:  Memory allocation failed.\n");
	SDL_Quit();
	exit(-1);
    }

    if (musicsource==1) {
	fprintf(stderr,"Opening music output...\n");
#ifdef WIN32
	if ((i=midiOutOpen(&sequencerdevice,MIDI_MAPPER,(DWORD)(NULL),
			   (DWORD)(NULL),0))!=
	    MMSYSERR_NOERROR) {
	    fprintf(stderr,"Failed to open MIDI Mapper; error %d.\n",i);
	    SDL_Quit();
	    exit(-1);
	}
#else
	sequencerdevice=open("/dev/sequencer", O_WRONLY, 0);
	if (sequencerdevice<0) {
	    fprintf(stderr,"Music failed opening /dev/sequencer.\n");
	    SDL_Quit();
	    exit(-1);
	}
	if (ioctl(sequencerdevice, SNDCTL_SEQ_NRMIDIS, &nrmidis) == -1) {
	    fprintf(stderr, "Can't get info about midi ports!\n");
	    SDL_Quit();
	    exit(-1);
	}
#endif
    }

    if (speechstatus >= 2)
    {
	SoundFile=malloc(SND_FILE_LENGTH);

	SoundBuffer=malloc(65536*2);

	if ((SoundBuffer==NULL)||(SoundBuffer==NULL)) {
	    fprintf(stderr,"Insufficient memory for sound.\n");
	    SDL_Quit();
	    exit(-1);
	}

	file=fopen("sounds.kzp","rb");
	if (file==NULL) {
	    fprintf(stderr,"Can not find sounds.kzp.\n");
	    SDL_Quit();
	    exit(-1);
	}
	if (fread(SoundFile,1,SND_FILE_LENGTH,file)!=SND_FILE_LENGTH) {
	    fprintf(stderr,"Error in sounds.kzp.\n");
	    SDL_Quit();
	    exit(-1);
	}
	fclose(file);

	SDL_LockMutex(soundmutex);
	fprintf(stderr,"Opening sound output...\n");
	want.freq=11025; /* This is probably the intended sample rate.*/
	want.format=AUDIO_S16;
	want.channels=1;
	want.samples=SOUNDBLOCKSIZE;
	want.userdata=NULL;
	want.callback=AudioCallback;	

	SDL_OpenAudio(&want,NULL);

	reset_dsp();

	SDL_UnlockMutex(soundmutex);
	SDL_PauseAudio(0);
    }
    fprintf(stderr,"Allocating memory...\n");
    if ((lzwbuf = convallocate(12304>>4)) == NULL)
    {
	fprintf(stderr,"Error #3:  Memory allocation failed.\n");
	SDL_Quit();
	exit(-1);
    }

    convwalls = numwalls;

    if ((pic = convallocate((numwalls-initialwalls)<<8)) == NULL)
    {
	fprintf(stderr,
		"Error #4: This computer does not have enough memory.\n");
	SDL_Quit();
	exit(-1);
    }
    walcounter = initialwalls;
    if (convwalls > initialwalls)
    {
	v = pic;
	for(i=0;i<convwalls-initialwalls;i++)
	{
	    walseg[walcounter] = v;
	    walcounter++;
	    v += 4096;
	}
    }
    l = 0;
    for(i=0;i<240;i++)
    {
	times90[i] = l;
	l += 90;
    }
    less64inc[0] = 16384;
    for(i=1;i<64;i++)
	less64inc[i] = 16384 / i;
    speed = 240;
    for(i=0;i<256;i++)
	keystatus[i] = 0;
    for(i=0;i<SDLKEYS;i++)
	newkeystatus[i]=0;

    if (largescreentexture) {
	glGenTextures(1,&screenbuffertexture);
    } else {
	glGenTextures(48,screenbuffertextures);
    }

    saidwelcome = 0;
    loadmusic("BEGIN");
    clockspeed = 0;
    slottime = 0;
    owecoins = 0;
    owecoinwait = 0;
    slotpos[0] = 0;
    slotpos[1] = 0;
    slotpos[2] = 0;
    clockspeed = 0;
    skilevel = 0;
    musicon();
    fprintf(stderr,"Loading intro pictures...\n");

    /* The ingame palette is stored in this GIF! */
    kgif(1);
    memcpy(spritepalette,palette,768);

    /* Show the Epic Megagames logo while loading... */

    kgif(0);
    fprintf(stderr,"Loading graphics...\n");
    loadwalls();

    /* Ken's Labyrinth logo. */
    kgif(1);

    fade(63);
  
    SetVisibleScreenOffset(0);
    SDL_GL_SwapBuffers();

    if (moustat == 0)
	moustat = setupmouse();
    SDL_LockMutex(timermutex);
    oclockspeed = clockspeed;
    while ((keystatus[1] == 0) && (keystatus[57] == 0) && 
	   (keystatus[28] == 0) && (bstatus == 0) &&
	   (clockspeed < oclockspeed+960))
    {
	PollInputs();

	bstatus = 0;
	if (moustat == 0)
	{
	    bstatus=readmouse(NULL, NULL);
	}
	if (joystat == 0)
	{
	    bstatus|=readjoystick(NULL,NULL);
	}
	SDL_UnlockMutex(timermutex);
	SDL_Delay(10);
	SDL_LockMutex(timermutex);
    }
    SDL_UnlockMutex(timermutex);

    /* Big scrolly picture... */

    fade(0);
    kgif(2);
    fade(63);
    l = 25200;
    i = 1;

    SDL_LockMutex(timermutex);
    oclockspeed=clockspeed;

    while ((keystatus[1] == 0) && (keystatus[57] == 0) &&
	   (keystatus[28] == 0) && (bstatus == 0) && (clockspeed < 7680))
    {
	if (i == 0)
	{
	    l += 90;
	    if (l >= 25200)
	    {
		i = 1;
		l = 25200;
	    }
	}
	if (i == 1)
	{
	    l -= 90;
	    if (l > 32768)
	    {
		l = 0;
		i = 0;
	    }
	}

	while(clockspeed<oclockspeed+12) {
	    SDL_UnlockMutex(timermutex);
	    SDL_Delay(10);
	    SDL_LockMutex(timermutex);
	}
	oclockspeed+=12;

	SDL_UnlockMutex(timermutex);
	glClearColor(0,0,0,0);
	glClear( GL_COLOR_BUFFER_BIT);
	visiblescreenyoffset=(l/90)-20;
	ShowPartialOverlay(20,20+visiblescreenyoffset,320,200);
	SDL_GL_SwapBuffers();
	SDL_LockMutex(timermutex);
    
	PollInputs();
	bstatus = 0;
	if (moustat == 0)
	{
	    bstatus=readmouse(NULL, NULL);
	}
	if (joystat == 0)
	{
	    bstatus|=readjoystick(NULL,NULL);
	}
    }
    oclockspeed=clockspeed;
    for(i=63;i>=0;i-=4)
    {
	SDL_UnlockMutex(timermutex);
	fade(64+i);
	glClearColor(0,0,0,0);
	glClear( GL_COLOR_BUFFER_BIT);
	visiblescreenyoffset=(l/90)-20;
	ShowPartialOverlay(20,20+visiblescreenyoffset,320,200);
	SDL_GL_SwapBuffers();
	SDL_LockMutex(timermutex);

	while(clockspeed<oclockspeed+4) {
	    SDL_UnlockMutex(timermutex);
	    SDL_Delay(10);
	    SDL_LockMutex(timermutex);
	}
	oclockspeed+=4;
    }
    SDL_UnlockMutex(timermutex);
    k = 0;
    for(i=0;i<16;i++)
	for(j=1;j<17;j++)
	{
	    palette[k++] = (paldef[i][0]*j)/17;
	    palette[k++] = (paldef[i][1]*j)/17;
	    palette[k++] = (paldef[i][2]*j)/17;
	}
    lastunlock = 1;
    lastshoot = 1;
    lastbarchange = 1;
    cheatenable = 0;
    hiscorenamstat = 0;

    /* Shareware/registered check... */

    if ((i = open("boards.kzp",O_RDWR|O_BINARY,S_IREAD)) != -1)
    {
	read(i,&boleng[0],30*4);
	numboards = 30;
	if ((boleng[40]|boleng[41]) == 0)
	    numboards = 20;
	if ((boleng[20]|boleng[21]) == 0)
	    numboards = 10;
	close(i);
    } else {
	fprintf(stderr,"boards.kzp not found.\n");
	SDL_Quit();
	exit(1);
    }
    musicoff();
}
